#![allow(unused)]
use std::string::String;
use logos::Logos;
use logos::Lexer;
use core::iter::Peekable;

pub use Token::*;

#[derive(Debug, Logos, PartialEq, Eq, Clone)]
#[logos(skip r"[ \t\n\f]+")]
pub enum Token {
    // SINGLE CHARACTER TOKENS
    #[token(";")]
    Semicolon,	// ;
    #[token("=")]
    Equal,	// =
    #[token("(")]
    LeftParen,	// (
    #[token(")")]
    RightParen, // )
    #[token("{")]
    LeftBrace,	// {
    #[token("}")]
    RightBrace, // }
    #[token(",")]
    Comma,	// ,
    #[token(".")]
    Dot,	// .
    #[token("-")]
    Minus,	// -
    #[token("+")]
    Plus,	// +
    #[token("/")]
    Slash,	// /
    #[token("*")]
    Star,	// *
    #[token("%")]
    Percent,	// %
    #[token("!")]
    Bang,	// !
    #[token(":")]
    Colon,	// :
    #[token("<")]
    Less,	// <
    #[token(">")]
    Greater,	// >
    #[token("|")]
    Pipe,       // |

    // KEYWORDS
    #[token("fnaf")]
    Fnaf,	// fnaf
    #[token("let")]
    Let,	// let
    #[token("global")]
    Global,     // global
    #[token("if")]
    If,		// if
    #[token("else")]
    Else,	// else
    #[token("while")]
    While,	// while
    #[token("elif")]
    Elif,	// elif
    #[token("return")]
    Return,	// return
    #[token("for")]
    For,	// for
    #[token("in")]
    In,		// in
    #[token("break")]
    Break,	// break
    #[token("continue")]
    Continue,	// continue

    // TWO CHARACTER TOKENS
    #[token("==")]
    EqualEqual,		// ==
    #[token("!=")]
    BangEqual,		// !=
    #[token("<=")]
    LessEqual,		// <=
    #[token(">=")]
    GreaterEqual,	// >=
    
    // LITERALS
    #[regex(r#"("[^"]*")|('[^']*')"#)]
    StringLiteral, // A string literal.
    #[regex("[0-9]+", |lex| lex.slice().parse().ok())]
    Number(u64),   // An integer.
    #[regex(r#"[^[0-9]^"^-^[ \t\n\f]^\.^=^(^)^{^}.^,^;^[+-/*%]]+[^"^-^=^\..^[ \t\n\f]^(^)^{^}^,^;^[+-/*%]]*"#, |lex| lex.slice().to_owned())]
    Identifier(String),  // An identifier.
    #[token("true")]
    True,	 // true
    #[token("false")]
    False,	 // false
    #[token("none")]
    Null,	 // none
}

pub fn lex_str(this: &str) -> Vec<(Token, &str)> {
    let mut buf = Vec::new();
    let mut lexer = Token::lexer(this);
    while let Some(Ok(token)) = lexer.next() {
	buf.push((token, lexer.slice()));
    }

    buf
}


impl std::fmt::Display for Token {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match self {
            Semicolon => write!(f, ";"),
            Equal => write!(f, "="),
            LeftParen => write!(f, "("),
            RightParen => write!(f, ")"),
            LeftBrace => write!(f, "{{"),
            RightBrace => write!(f, "}}"),
            Comma => write!(f, ","),
            Dot => write!(f, "."),
            Minus => write!(f, "-"),
            Plus => write!(f, "+"),
            Slash => write!(f, "/"),
            Star => write!(f, "*"),
            Percent => write!(f, "%"),
            Bang => write!(f, "!"),
            Colon => write!(f, ":"),
            Less => write!(f, "<"),
            Greater => write!(f, ">"),
            Pipe => write!(f, "|"),
            Fnaf => write!(f, "fnaf"),
            Let => write!(f, "let"),
            Global => write!(f, "global"),
            If => write!(f, "if"),
            Else => write!(f, "else"),
            While => write!(f, "while"),
            Elif => write!(f, "elif"),
            Return => write!(f, "return"),
            For => write!(f, "for"),
            In => write!(f, "in"),
            Break => write!(f, "break"),
            Continue => write!(f, "continue"),
            EqualEqual => write!(f, "=="),
            BangEqual => write!(f, "!="),
            LessEqual => write!(f, "<="),
            GreaterEqual => write!(f, ">="),
            StringLiteral => write!(f, "string"),
            Number(n) => write!(f, "{}", n),
            Identifier(i) => write!(f, "{}", i),
            True => write!(f, "true"),
            False => write!(f, "false"),
            Null => write!(f, "none"),
        }
    }
}
